@@ -38,6 +38,7 @@ correctly and Exception-safe
 #include <cmath>
 #include <string.h>
 #include <vector>
+#include <charconv>
 
 namespace NMR {
 
@@ -153,6 +154,9 @@ namespace NMR {
 		__NMRASSERT(pwszValue);
 		nfDouble dResult = 0.0;
 
+		//skip leading whitespaces
+		for (; *pszValue && *pszValue == ' '; pszValue++);
+
 		// Convert to double and make a input and range check!
 		auto answer = fast_float::from_chars(pszValue, pszValue + strlen(pszValue), dResult);
 
